@@ -7,8 +7,8 @@ searching etc. See :ref:`walking` for details.
 
 from __future__ import unicode_literals
 
-import itertools
-
+import typing
+from collections import defaultdict
 from collections import deque
 from collections import namedtuple
 
@@ -20,118 +20,59 @@ from .path import abspath
 from .path import join
 from .path import normpath
 
+if False:  # typing.TYPE_CHECKING
+    from typing import (
+        Any, Callable, Collection, Iterator, List,
+        Optional, MutableMapping, Text, Tuple, Type)
+    from .base import FS
+    from .info import Info
+    OnError = Callable[[Text, Exception], bool]
+
+
+_F = typing.TypeVar('_F', bound='FS')
+
 
 Step = namedtuple('Step', 'path, dirs, files')
 """type: a *step* in a directory walk.
 """
 
 
-class WalkerBase(object):
-    """The base class for a Walker.
+# TODO(@althonos): It could be a good idea to create an Abstract Base Class
+#                  BaseWalker (with methods walk, files, dirs and info) ?
 
-    To create a custom walker, implement `~fs.walk.WalkerBase.walk`
-    in a sub-class.
-
-    See `~fs.walk.Walker` for a fully featured walker object that
-    should be adequate for all but your most exotic directory walking
-    needs.
-
-    """
-
-    def walk(self, fs, path='/', namespaces=None):
-        """Walk the directory structure of a filesystem.
-
-        Arguments:
-            fs (FS): A filesystem instance.
-            path (str, optional): A path to a directory on the filesystem.
-            namespaces(list, optional): A list of additional namespaces
-                to add to the `~fs.info.Info` objects.
-
-        Returns:
-            ~collections.Iterator: iterator of `~fs.walk.Step` named tuples.
-
-        """
-        raise NotImplementedError
-
-    def files(self, fs, path='/'):
-        """Walk a filesystem, yielding absolute paths to files.
-
-        Arguments:
-            fs (FS): A filesystem instance.
-            path (str, optional): A path to a directory on the filesystem.
-
-        Yields:
-            str: absolute path to files on the filesystem found
-            recursively within the given directory.
-
-        """
-        iter_walk = iter(self.walk(fs, path=path))
-        for _path, _, files in iter_walk:
-            for info in files:
-                yield join(_path, info.name)
-
-    def dirs(self, fs, path='/'):
-        """Walk a filesystem, yielding absolute paths to directories.
-
-        Arguments:
-            fs (FS): A filesystem instance.
-            path (str, optional): A path to a directory on the filesystem.
-
-        Yields:
-            str: absolute path to directories on the filesystem found
-            recursively within the given directory.
-
-        """
-        for path, dirs, _ in self.walk(fs, path=path):
-            for info in dirs:
-                yield join(path, info.name)
-
-    def info(self, fs, path='/', namespaces=None):
-        """Walk a filesystem, yielding tuples of ``(<path>, <info>)``.
-
-        Arguments:
-            fs (FS): A filesystem instance.
-            path (str, optional): A path to a directory on the filesystem.
-            namespaces (list, optional): A list of additional namespaces
-                to add to the `Info` objects.
-
-        Yields:
-            (str, Info): a tuple of ``(<absolute path>, <resource info>)``.
-
-        """
-        _walk = self.walk(fs, path=path, namespaces=namespaces)
-        for _path, dirs, files in _walk:
-            for info in itertools.chain(dirs, files):
-                yield join(_path, info.name), info
-
-
-class Walker(WalkerBase):
+class Walker(object):
     """A walker object recursively lists directories in a filesystem.
 
     Arguments:
-        ignore_errors (bool, optional): If `True`, any errors reading
-            a directory will be ignored, otherwise exceptions will be
-            raised.
+        ignore_errors (bool): If `True`, any errors reading a
+            directory will be ignored, otherwise exceptions will
+            be raised.
         on_error (callable, optional): If ``ignore_errors`` is `False`,
             then this callable will be invoked for a path and the exception
             object. It should return `True` to ignore the error, or `False`
             to re-raise it.
-        search (str, optional): If ``'breadth'`` then the directory will be
+        search (str): If ``'breadth'`` then the directory will be
             walked *top down*. Set to ``'depth'`` to walk *bottom up*.
-        filter (list, optional): If supplied, this parameter should be a
-            list of filename patterns, e.g. ``['*.py']``. Files will only
-            be returned if the final component matches one of the patterns.
-        exclude_dirs (list, optional): A list of patterns that will be used
-            to filter out directories from the walk. e.g. ``['*.svn', '*.git']``.
+        filter (list, optional): If supplied, this parameter should be
+            a list of filename patterns, e.g. ``['*.py']``. Files will
+            only be returned if the final component matches one of the
+            patterns.
+        exclude_dirs (list, optional): A list of patterns that will be
+            used to filter out directories from the walk. e.g.
+            ``['*.svn', '*.git']``.
+        max_depth (int, optional): Maximum directory depth to walk.
 
     """
 
     def __init__(self,
-                 ignore_errors=False,
-                 on_error=None,
-                 search="breadth",
-                 filter=None,
-                 exclude_dirs=None):
+                 ignore_errors=False,       # type: bool
+                 on_error=None,             # type: Optional[OnError]
+                 search="breadth",          # type: Text
+                 filter=None,               # type: Optional[List[Text]]
+                 exclude_dirs=None,         # type: Optional[List[Text]]
+                 max_depth=None             # type: Optional[int]
+                 ):
+        # type: (...) -> None
         if search not in ('breadth', 'depth'):
             raise ValueError("search must be 'breadth' or 'depth'")
         self.ignore_errors = ignore_errors
@@ -153,20 +94,33 @@ class Walker(WalkerBase):
         self.search = search
         self.filter = filter
         self.exclude_dirs = exclude_dirs
+        self.max_depth = max_depth
         super(Walker, self).__init__()
 
     @classmethod
     def _ignore_errors(cls, path, error):
+        # type: (Text, Exception) -> bool
         """Default on_error callback."""
         return True
 
     @classmethod
     def _raise_errors(cls, path, error):
+        # type: (Text, Exception) -> bool
         """Callback to re-raise dir scan errors."""
         return False
 
+    @classmethod
+    def _calculate_depth(cls, path):
+        # type: (Text) -> int
+        """Calculate the 'depth' of a directory path (number of
+        components).
+        """
+        _path = path.strip('/')
+        return _path.count('/') + 1 if _path else 0
+
     @classmethod
     def bind(cls, fs):
+        # type: (_F) -> BoundWalker[_F]
         """Bind a `Walker` instance to a given filesystem.
 
         This *binds* in instance of the Walker to a given filesystem, so
@@ -202,55 +156,83 @@ class Walker(WalkerBase):
         return BoundWalker(fs)
 
     def __repr__(self):
+        # type: () -> Text
         return make_repr(
             self.__class__.__name__,
             ignore_errors=(self.ignore_errors, False),
             on_error=(self.on_error, None),
             search=(self.search, 'breadth'),
             filter=(self.filter, None),
-            exclude_dirs=(self.exclude_dirs, None)
+            exclude_dirs=(self.exclude_dirs, None),
+            max_depth=(self.max_depth, None)
         )
 
-    def filter_files(self, fs, infos):
-        """Filter a sequence of resource `Info` objects.
-
-        The default implementation filters those files for which
-        `~fs.walk.Walker.check_file` returns `True`.
-
-        Arguments:
-            fs (FS): a filesystem instance.
-            infos (list): A list of `~fs.info.Info` instances.
-
-        Returns:
-            list: a list of `Info` objects passing the ``check_file``
-            validation.
+    def _iter_walk(self,
+                   fs,                # type: FS
+                   path,              # type: Text
+                   namespaces=None    # type: Optional[Collection[Text]]
+                   ):
+        # type: (...) -> Iterator[Tuple[Text, Optional[Info]]]
+        """Get the walk generator."""
+        if self.search == 'breadth':
+            return self._walk_breadth(fs, path, namespaces=namespaces)
+        else:
+            return self._walk_depth(fs, path, namespaces=namespaces)
 
+    def _check_open_dir(self, fs, path, info):
+        # type: (FS, Text, Info) -> bool
+        """Check if a directory should be considered in the walk.
         """
-        _check_file = self.check_file
-        return [
-            info for info in infos
-            if _check_file(fs, info)
-        ]
+        if (self.exclude_dirs is not None and
+                fs.match(self.exclude_dirs, info.name)):
+            return False
+        return self.check_open_dir(fs, path, info)
 
-
-    def check_open_dir(self, fs, info):
+    def check_open_dir(self, fs, path, info):
+        # type: (FS, Text, Info) -> bool
         """Check if a directory should be opened.
 
         Override to exclude directories from the walk.
 
         Arguments:
             fs (FS): A filesystem instance.
-            info (Info): A resource info object.
+            path (str): Path to directory.
+            info (Info): A resource info object for the directory.
 
         Returns:
             bool: `True` if the directory should be opened.
 
         """
-        if self.exclude_dirs is None:
-            return True
-        return not fs.match(self.exclude_dirs, info.name)
+        return True
+
+    def _check_scan_dir(self, fs, path, info, depth):
+        # type: (FS, Text, Info, int) -> bool
+        """Check if a directory contents should be scanned."""
+        if self.max_depth is not None and depth >= self.max_depth:
+            return False
+        return self.check_scan_dir(fs, path, info)
+
+    def check_scan_dir(self, fs, path, info):
+        # type: (FS, Text, Info) -> bool
+        """Check if a directory should be scanned.
+
+        Override to omit scanning of certain directories. If a directory
+        is omitted, it will appear in the walk but its files and
+        sub-directories will not.
+
+        Arguments:
+            fs (FS): A filesystem instance.
+            path (str): Path to directory.
+            info (Info): A resource info object for the directory.
+
+        Returns:
+            bool: `True` if the directory should be scanned.
+
+        """
+        return True
 
     def check_file(self, fs, info):
+        # type: (FS, Info) -> bool
         """Check if a filename should be included.
 
         Override to exclude files from the walk.
@@ -265,7 +247,12 @@ class Walker(WalkerBase):
         """
         return fs.match(self.filter, info.name)
 
-    def _scan(self, fs, dir_path, namespaces):
+    def _scan(self,
+              fs,               # type: FS
+              dir_path,         # type: Text
+              namespaces=None   # type: Optional[Collection[Text]]
+              ):
+        # type: (...) -> Iterator[Info]
         """Get an iterator of `Info` objects for a directory path.
 
         Arguments:
@@ -286,12 +273,17 @@ class Walker(WalkerBase):
             if not self.on_error(dir_path, error):
                 six.reraise(type(error), error)
 
-    def walk(self, fs, path='/', namespaces=None):
+    def walk(self,
+             fs,                # type: FS
+             path='/',          # type: Text
+             namespaces=None    # type: Optional[Collection[Text]]
+             ):
+        # type: (...) -> Iterator[Step]
         """Walk the directory structure of a filesystem.
 
         Arguments:
             fs (FS): A filesystem instance.
-            path (str, optional): A path to a directory on the filesystem.
+            path (str): A path to a directory on the filesystem.
             namespaces (list, optional): A list of additional namespaces
                 to add to the `Info` objects.
 
@@ -315,80 +307,149 @@ class Walker(WalkerBase):
 
         """
         _path = abspath(normpath(path))
+        dir_info = defaultdict(list)  # type: MutableMapping[Text, List[Info]]
+        _walk = self._iter_walk(fs, _path, namespaces=namespaces)
+        for dir_path, info in _walk:
+            if info is None:
+                dirs = []       # type: List[Info]
+                files = []      # type: List[Info]
+                for _info in dir_info[dir_path]:
+                    (dirs if _info.is_dir else files).append(_info)
+                yield Step(dir_path, dirs, files)
+                del dir_info[dir_path]
+            else:
+                dir_info[dir_path].append(info)
 
-        if self.search == 'breadth':
-            do_walk = self._walk_breadth
-        elif self.search == 'depth':
-            do_walk = self._walk_depth
+    def files(self, fs, path='/'):
+        # type: (FS, Text) -> Iterator[Text]
+        """Walk a filesystem, yielding absolute paths to files.
 
-        return do_walk(fs, _path, namespaces=namespaces)
+        Arguments:
+            fs (FS): A filesystem instance.
+            path (str): A path to a directory on the filesystem.
 
-    def _walk_breadth(self, fs, path, namespaces=None):
+        Yields:
+            str: absolute path to files on the filesystem found
+            recursively within the given directory.
+
+        """
+        for _path, info in self._iter_walk(fs, path=path):
+            if info is not None and not info.is_dir:
+                yield join(_path, info.name)
+
+    def dirs(self, fs, path='/'):
+        # type: (FS, Text) -> Iterator[Text]
+        """Walk a filesystem, yielding absolute paths to directories.
+
+        Arguments:
+            fs (FS): A filesystem instance.
+            path (str): A path to a directory on the filesystem.
+
+        Yields:
+            str: absolute path to directories on the filesystem found
+            recursively within the given directory.
+
+        """
+        for _path, info in self._iter_walk(fs, path=path):
+            if info is not None and info.is_dir:
+                yield join(_path, info.name)
+
+    def info(self,
+             fs,              # type: FS
+             path='/',        # type: Text
+             namespaces=None  # type: Optional[Collection[Text]]
+             ):
+        # type: (...) -> Iterator[Tuple[Text, Info]]
+        """Walk a filesystem, yielding tuples of ``(<path>, <info>)``.
+
+        Arguments:
+            fs (FS): A filesystem instance.
+            path (str): A path to a directory on the filesystem.
+            namespaces (list, optional): A list of additional namespaces
+                to add to the `Info` objects.
+
+        Yields:
+            (str, Info): a tuple of ``(<absolute path>, <resource info>)``.
+
+        """
+        _walk = self._iter_walk(fs, path=path, namespaces=namespaces)
+        for _path, info in _walk:
+            if info is not None:
+                yield join(_path, info.name), info
+
+    def _walk_breadth(self,
+                      fs,               # type: FS
+                      path,             # type: Text
+                      namespaces=None   # type: Optional[Collection[Text]]
+                      ):
+        # type: (...) -> Iterator[Tuple[Text, Optional[Info]]]
         """Walk files using a *breadth first* search.
         """
         queue = deque([path])
         push = queue.appendleft
         pop = queue.pop
+        depth = self._calculate_depth(path)
 
         while queue:
             dir_path = pop()
-            dirs = []
-            files = []
             for info in self._scan(fs, dir_path, namespaces=namespaces):
                 if info.is_dir:
-                    if self.check_open_dir(fs, info):
-                        dirs.append(info)
-                        push(join(dir_path, info.name))
+                    _depth = self._calculate_depth(dir_path) - depth + 1
+                    if self._check_open_dir(fs, dir_path, info):
+                        yield dir_path, info  # Opened a directory
+                        if self._check_scan_dir(fs, dir_path, info, _depth):
+                            push(join(dir_path, info.name))
                 else:
-                    files.append(info)
-            yield (
-                dir_path,
-                dirs,
-                self.filter_files(fs, files)
-            )
+                    if self.check_file(fs, info):
+                        yield dir_path, info  # Found a file
+            yield dir_path, None  # End of directory
 
-    def _walk_depth(self, fs, path, namespaces=None):
+    def _walk_depth(self,
+                    fs,               # type: FS
+                    path,             # type: Text
+                    namespaces=None   # type: Optional[Collection[Text]]
+                    ):
+        # type: (...) -> Iterator[Tuple[Text, Optional[Info]]]
         """Walk files using a *depth first* search.
         """
         # No recursion!
 
         def scan(path):
+            # type: (Text) -> Iterator[Info]
+            """Perform scan."""
             return self._scan(fs, path, namespaces=namespaces)
 
-        stack = [(
-            path, scan(path), [], []
-        )]
+        stack = [(path, scan(path), None)]  # type: List[Tuple[Text, Iterator[Info], Optional[Tuple[Text, Info]]]]
+        depth = self._calculate_depth(path)
         push = stack.append
 
         while stack:
-            dir_path, iter_files, dirs, files = stack[-1]
-            try:
-                info = next(iter_files)
-            except StopIteration:
-                yield (
-                    dir_path,
-                    dirs,
-                    self.filter_files(fs, files)
-                )
+            dir_path, iter_files, parent = stack[-1]
+            info = next(iter_files, None)
+            if info is None:
+                if parent is not None:
+                    yield parent
+                yield dir_path, None
                 del stack[-1]
-            else:
-                if info.is_dir:
-                    if self.check_open_dir(fs, info):
-                        dirs.append(info)
+            elif info.is_dir:
+                _depth = self._calculate_depth(dir_path) - depth + 1
+                if self._check_open_dir(fs, dir_path, info):
+                    if self._check_scan_dir(fs, dir_path, info, _depth):
                         _path = join(dir_path, info.name)
-                        push((
-                            _path, scan(_path), [], []
-                        ))
-                else:
-                    files.append(info)
+                        push((_path, scan(_path), (dir_path, info)))
+                    else:
+                        yield dir_path, info
+            else:
+                if self.check_file(fs, info):
+                    yield dir_path, info
 
 
-class BoundWalker(object):
+class BoundWalker(typing.Generic[_F]):
     """A class that binds a `Walker` instance to a `FS` instance.
 
     Arguments:
         fs (FS): A filesystem instance.
-        walker_class (type, optional): A `~fs.walk.WalkerBase`
+        walker_class (type): A `~fs.walk.WalkerBase`
             sub-class. The default uses `~fs.walk.Walker`.
 
     You will typically not need to create instances of this class
@@ -407,26 +468,31 @@ class BoundWalker(object):
     """
 
     def __init__(self, fs, walker_class=Walker):
+        # type: (_F, Type[Walker]) -> None
         self.fs = fs
         self.walker_class = walker_class
 
     def __repr__(self):
+        # type: () -> Text
         return "BoundWalker({!r})".format(self.fs)
 
     def _make_walker(self, *args, **kwargs):
+        # type: (*Any, **Any) -> Walker
         """Create a walker instance.
         """
         walker = self.walker_class(*args, **kwargs)
         return walker
 
     def walk(self,
-             path='/',
-             namespaces=None,
-             **kwargs):
+             path='/',          # type: Text
+             namespaces=None,   # type: Optional[Collection[Text]]
+             **kwargs           # type: Any
+             ):
+        # type: (...) -> Iterator[Step]
         """Walk the directory structure of a filesystem.
 
         Arguments:
-            path (str, optional):
+            path (str):
             namespaces (list, optional): A list of namespaces to include
                 in the resource information, e.g. ``['basic', 'access']``
                 (defaults to ``['basic']``).
@@ -448,6 +514,7 @@ class BoundWalker(object):
             exclude_dirs (list): A list of patterns that will be used
                 to filter out directories from the walk, e.g. ``['*.svn',
                 '*.git']``.
+            max_depth (int, optional): Maximum directory depth to walk.
 
         Returns:
             ~collections.Iterator: an iterator of ``(<path>, <dirs>, <files>)``
@@ -458,7 +525,7 @@ class BoundWalker(object):
         Example:
             >>> home_fs = open_fs('~/')
             >>> walker = Walker(filter=['*.py'])
-            >>> for path, dirs, files in walker.walk(home_fs, ['details']):
+            >>> for path, dirs, files in walker.walk(home_fs, namespaces=['details']):
             ...     print("[{}]".format(path))
             ...     print("{} directories".format(len(dirs)))
             ...     total = sum(info.size for info in files)
@@ -473,6 +540,7 @@ class BoundWalker(object):
     __call__ = walk
 
     def files(self, path='/', **kwargs):
+        # type: (Text, **Any) -> Iterator[Text]
         """Walk a filesystem, yielding absolute paths to files.
 
         Arguments:
@@ -495,9 +563,10 @@ class BoundWalker(object):
             exclude_dirs (list): A list of patterns that will be used
                 to filter out directories from the walk, e.g. ``['*.svn',
                 '*.git']``.
+            max_depth (int, optional): Maximum directory depth to walk.
 
         Returns:
-            ~collections.Iterable: An iterable of file paths (absolute
+            ~collections.Iterator: An iterator over file paths (absolute
             from the filesystem root).
 
         This method invokes `Walker.files` with the bound `FS` object.
@@ -507,6 +576,7 @@ class BoundWalker(object):
         return walker.files(self.fs, path=path)
 
     def dirs(self, path='/', **kwargs):
+        # type: (Text, **Any) -> Iterator[Text]
         """Walk a filesystem, yielding absolute paths to directories.
 
         Arguments:
@@ -525,9 +595,10 @@ class BoundWalker(object):
             exclude_dirs (list): A list of patterns that will be used
                 to filter out directories from the walk, e.g. ``['*.svn',
                 '*.git']``.
+            max_depth (int, optional): Maximum directory depth to walk.
 
         Returns:
-            ~collections.iterable: an iterable of directory paths
+            ~collections.Iterator: an iterator over directory paths
             (absolute from the filesystem root).
 
         This method invokes `Walker.dirs` with the bound `FS` object.
@@ -536,11 +607,16 @@ class BoundWalker(object):
         walker = self._make_walker(**kwargs)
         return walker.dirs(self.fs, path=path)
 
-    def info(self, path='/', namespaces=None, **kwargs):
+    def info(self,
+             path='/',          # type: Text
+             namespaces=None,   # type: Optional[Collection[Text]]
+             **kwargs           # type: Any
+             ):
+        # type: (...) -> Iterator[Tuple[Text, Info]]
         """Walk a filesystem, yielding path and `Info` of resources.
 
         Arguments:
-            path (str, optional):
+            path (str): A path to a directory.
             namespaces (list, optional): A list of namespaces to include
                 in the resource information, e.g. ``['basic', 'access']``
                 (defaults to ``['basic']``).
@@ -562,6 +638,7 @@ class BoundWalker(object):
             exclude_dirs (list): A list of patterns that will be used
                 to filter out directories from the walk, e.g. ``['*.svn',
                 '*.git']``.
+            max_depth (int, optional): Maximum directory depth to walk.
 
         Returns:
             ~collections.Iterable: an iterable yielding tuples of
